#include "lower_detect_limit.h"
#include "common.h"
#include "ldl_logistic.h"
#include <iostream>
#include <vector>
#include <string>
#include <eigen3/Eigen/Dense>
using std::cout;
using std::endl;
using std::string;
using std::vector;
using namespace common;

void write_full_estimate_header( std::ofstream & fullest_out_file,
                                 bool pheno_covar_snp,
                                 bool isPhenoBinary,
                                 const vector< string > & covar_name_vec )
{
             fullest_out_file << "SNP" << "\t";
            //estimates beta
            fullest_out_file << "beta_constant" << "\t";
            if ( pheno_covar_snp )
            {
                fullest_out_file << "beta_G" << "\t";
            }
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "beta_" << covar_name 
                               << "\t";
            }
            //se beta
            fullest_out_file << "beta_constant_se" << "\t";
            if ( pheno_covar_snp )
            {
                fullest_out_file << "beta_G_se" << "\t";
            }
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "beta_" << covar_name
                               << "_se" << "\t";
            }
            //p-val beta
            fullest_out_file << "beta_constant_pval" << "\t";
            if ( pheno_covar_snp )
            {
                fullest_out_file << "beta_G_pval" << "\t";
            }
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "beta_" << covar_name
                               << "_pval" << "\t";
            }
            fullest_out_file << "gamma" << "\t";
            fullest_out_file << "gamma_se" << "\t";
            fullest_out_file << "gamma_pval" << "\t";
            if ( false == isPhenoBinary )
            {
                fullest_out_file << "sigma_sqr" << "\t";
                fullest_out_file << "sigma_sqr_se" << "\t";
                fullest_out_file << "sigma_sqr_pval" << "\t";
            }
            //alpha estimate
            fullest_out_file << "alpha_constant" << "\t";
            fullest_out_file << "alpha_G" << "\t";
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "alpha_" << covar_name 
                               << "\t";
            }
            //alpha se
            fullest_out_file << "alpha_constant_se" << "\t";
            fullest_out_file << "alpha_G_se" << "\t";
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "alpha_" << covar_name
                               << "_se" << "\t";
            }
            //alpha pval
            fullest_out_file << "alpha_constant_pval" << "\t";         
            fullest_out_file << "alpha_G_pval" << "\t";
            for ( const auto & covar_name : covar_name_vec )
            {
                fullest_out_file << "alpha_" << covar_name
                               << "_pval" << "\t";
            }
            fullest_out_file << "tau_sqr" << "\t";
            fullest_out_file << "tau_sqr_se" << "\t";
            fullest_out_file << "tau_sqr_pval" << "\n";
}

void write_full_estimate_data( std::ofstream & fullest_out_file,
                               const VectorXd & full_estimate_vec,
                               const VectorXd & se_vec,
                               size_t NUM_COOR_X,
                               size_t NUM_COOR_Z,
                               double estimate_gamma,
                               double estimate_sigma_sqr,
                               double estimate_tau_sqr,
                               bool isPhenoBinary )
{
            VectorXd full_est_z_stat; 
            calc_z_stat_vec( full_estimate_vec,
                         se_vec,
                         full_est_z_stat );
            VectorXd full_est_pval;
            calc_p_val_vec( full_est_z_stat, full_est_pval );
            //beta estimate             
            for ( size_t index = 0; index < NUM_COOR_X; ++index )
            {
                 fullest_out_file << full_estimate_vec( index ) << "\t";
            }
            //beta se
            for ( size_t index = 0; index < NUM_COOR_X; ++index )
            {
                 fullest_out_file << se_vec( index ) << "\t";
            }
            //beta pval 
            for ( size_t index = 0; index < NUM_COOR_X; ++index )            
            {
                 fullest_out_file << full_est_pval( index ) << "\t";           
            }
            //gamma est
            fullest_out_file << estimate_gamma << "\t";
            //gamma se
            fullest_out_file << se_vec( NUM_COOR_X ) << "\t";
            //gamma_pval
            fullest_out_file << full_est_pval( NUM_COOR_X ) << "\t";
            if ( false == isPhenoBinary )
            {
                //sigma sqr est
                fullest_out_file << estimate_sigma_sqr << "\t";
                //sigma sqr se
                fullest_out_file << se_vec( NUM_COOR_X  + 1 ) << "\t";
                //sigma_sqr_pval
                fullest_out_file << full_est_pval( NUM_COOR_X  + 1 ) << "\t";
            }

            const size_t KSI_FIRST = (isPhenoBinary)? 
                         NUM_COOR_X + 1 : NUM_COOR_X + 2;
            const size_t LAST_INDEX = KSI_FIRST + NUM_COOR_Z;
            //cout << "KSI_FIRST == " << KSI_FIRST << endl;
            //cout << "NUM_COOR_Z == " << NUM_COOR_Z << endl;
            //alpha est
            for ( size_t index = KSI_FIRST; index < LAST_INDEX; ++index )
            {
                 fullest_out_file << full_estimate_vec( index ) << "\t";
            }
            //alpha se
            for ( size_t index = KSI_FIRST; index < LAST_INDEX; ++index )
            {
                 fullest_out_file << se_vec( index ) << "\t";
            }
            //alpha pval
            for ( size_t index = KSI_FIRST; index < LAST_INDEX; ++index )
            {
                 fullest_out_file << full_est_pval( index ) << "\t";
            }
            const size_t TAU2_INDEX = full_estimate_vec.rows() - 1;
            //tau sqr est
            fullest_out_file << estimate_tau_sqr << "\t";
            //tau sqr se
            fullest_out_file << se_vec( TAU2_INDEX ) << "\t";
            //tau sqr pval
            fullest_out_file << full_est_pval( TAU2_INDEX );
            fullest_out_file << "\n";
}                               

void write_info_message( const string & exe_name )
{
        cout << "Usage: " << exe_name <<
            " config_filename" << endl;
        cout << "where config_filename is the name ";
        cout << "of a filename containing settings ";
        cout << endl << "for the calculations." << endl << endl;
        cout << "The following settings are recognized ";
        cout << "in a configration file:";
        cout << endl;
        cout << "geno_filename geno_filename_string" << endl;
        cout << "where geno_filename_string is the name of the ";
        cout << "file with genotype data. (required)" << endl << endl;
        cout << "pheno_filename pheno_filename_string" << endl;
        cout << "where pheno_filename_string is tbe name of the ";
        cout << "file containing phenotype, " << endl;
        cout << "quantitative measure subject to detection limit, ";
        cout << "and covariate data." << endl;
        cout << "The phenotype values should be numeric for ";
        cout << "non-missing values (missing phenotype should be 'NA'). ";
        cout << endl;
        cout << "Binary phenotypes should be coded as 0 or 1 for ";
        cout << "non-missing values. ";
        cout << " (required)" << endl << endl;
        cout << "pheno_covar_snp boolean_string where boolean_string ";
        cout << "is either 'true' or 'false'." << endl;
        cout << "If set to true then genotype is included in ";
        cout << "X matrix as defined in documentation (required)" << endl;
        cout << endl << endl;
        cout << "out_freq an_integer" << endl << 
                "where an_integer is a positive integer.";
        cout << endl << "Informational message frequency ";
        cout << " (optional)" << endl;
        cout << endl << "min_call_rate a_float where 0<= a_float <=1 ";
        cout << endl << "The minimum call rate for a SNP to be considered (optional)";
        cout << endl << endl << "out_suffix a_string where a_string is appended to the ";
        cout << endl << "phenotype file name to create the output file name (optional)";
        cout << endl << endl;
        cout << "write_covariance boolean_string where boolean_string ";
        cout << "is either 'true' or 'false'." << endl;
        cout << "If write_covariance is true then the full covariance ";
        cout << "matrix will be written to disk for each SNP.  By default ";
        cout << "write_covariance is false so that the covariance matrix ";
        cout << "will not be written (optional)" << endl << endl;
        cout << "write_full_estimate boolean_string where boolean_string ";
        cout << "is either 'true' or 'false'." << endl;
        cout << "If write_full_estimate is true then the full estimate ";
        cout << "vector (including all parameters) will be written to disk ";
        cout << "for each SNP. By default ";
        cout << "write_full_estimate is false so that the full estimate ";             
        cout << "vector will not be written (optional)" << endl << endl; 
    
        cout << "For more information consult the program documentation.";
        cout << endl << "**************************" << std::endl;
}

int main( int argc, char *argv[] )
{
    
    //cout << "Main of lower_detect_limit_snp_loop.cpp " << endl;
    cout << "**************************\n" << std::endl;
    if (argc != 2)
    {
        write_info_message( argv[0] );
        return EXIT_FAILURE;
    }
    std::stringstream convert_config_filename(argv[1]);
    string config_filename;
    if (!(convert_config_filename >> config_filename))
    {
        cout << "Could not convert command line ";
        cout << "argument to string type for required configuration filename" << endl;
        return EXIT_FAILURE;
    }
    string geno_filename;
    bool pheno_covar_snp;
    string pheno_filename;
    size_t out_freq;
    double min_call_rate;
    size_t max_iter;
    double converged_ratio;
    string out_suffix;
    bool write_covariance;
    bool write_full_estimate;
    const int read_configfile_rv = 
            read_configfile( config_filename,
                             geno_filename,
                             pheno_filename,
                             pheno_covar_snp,
                             out_freq,
                             min_call_rate,
                             max_iter,
                             converged_ratio,
                             out_suffix,
                             write_covariance,
                             write_full_estimate );
    if ( NO_ERROR != read_configfile_rv )
    {
       string my_error_str;
       int_to_error_string( read_configfile_rv, my_error_str );
       cout << "error : could not read configuration file " <<
               config_filename << " : error of type :  " 
            << my_error_str << endl;
       return EXIT_FAILURE;
    }
    else
    {
      cout << "Finished reading " << config_filename << 
              std::endl;
    }
    cout << endl;
    cout << "The following settings  will be used : " <<
            std::endl; 
    cout << "geno_filename " << geno_filename << endl;
    cout << "pheno_filename " << pheno_filename << endl;
    cout << "pheno_covar_snp " << std::boolalpha << pheno_covar_snp << endl;
    cout << "out_freq " << out_freq << endl;
    cout << "min_call_rate " << min_call_rate << endl;
    //cout << "converged_ratio " << converged_ratio << endl;
    cout << "out_suffix " << out_suffix << endl; 
    cout << "pheno_filename " << pheno_filename << endl;
    cout << "write_covariance " << std::boolalpha << write_covariance << endl;
    cout << "write_full_estimate " << std::boolalpha << write_full_estimate << endl; 
    cout << endl;
    //read pheno
    vector< measurement > biomarker_measurements;
    VectorXd Y_pheno_vxd;
    MatrixXd covar_mxd;
    std::vector<subject_data_type> subject_data_type_vec;
    std::vector< std::string > covar_name_vec;
    bool isPhenoBinary;
    int rv = read_pheno_and_biomarker( 
             pheno_filename, biomarker_measurements, Y_pheno_vxd, covar_mxd,
             subject_data_type_vec,
             covar_name_vec,
             isPhenoBinary );
    if ( rv != NO_ERROR )
    {
        string my_error_str;
        int_to_error_string( rv, my_error_str );
        cout << "Error : could not read " << pheno_filename 
             << "error of type : " << 
             my_error_str << endl;
        return EXIT_FAILURE; 
    }
    else
    {
       cout << "Finished reading from ";
       cout << pheno_filename << endl;
       if ( false == isPhenoBinary )
       {
           cout << "Continuous phenotype detected." << endl;
       }
       else
       {
           cout << "Binary phenotype detected." << endl;
       }
    }
    cout << "covariate names == " << endl;
    for ( const auto & covar_name : covar_name_vec )
    {
       cout << covar_name << endl;
    }
    cout << "Number of subjects : " << Y_pheno_vxd.rows();
    cout << endl;
    cout << "Number of covariates : " << covar_mxd.rows();
    cout << endl;
    vector< string > snp_name_vector;
    std::vector< VectorXd > snp_vxd_vec;
    rv = read_geno( geno_filename, snp_vxd_vec, snp_name_vector, 
                    subject_data_type_vec, min_call_rate );
    if ( rv != NO_ERROR ) 
    {
        string my_error_str;
        int_to_error_string( rv, my_error_str );
        cout << "error : could not read " << geno_filename;
       
        cout << " : error of type : " << my_error_str << endl;
        return EXIT_FAILURE;
    }
    else
    {
        cout << "Finished reading from " << geno_filename;
        cout << endl;
    }
    cout << "Number of subjects from genotype file: ";
    cout << snp_vxd_vec[0].size() << endl;
    const size_t NUM_SNPS = snp_vxd_vec.size();
    cout << "Number of SNPs in genotype file: " << NUM_SNPS 
         << std::endl;
    MatrixXd X_mat;
    if ( ! pheno_covar_snp )
    {
        make_mxd_with_constant_from_covar_mxd( covar_mxd, X_mat );
    }
 
    const string out_filename = pheno_filename + out_suffix;
    std::ofstream outfile( out_filename );
    const string covarmat_out_filename = out_filename + ".covarmat.txt";
    const string fullest_out_filename = out_filename + ".fullest.txt";
    std::ofstream covarmat_out_file;
    if ( write_covariance )
    {
        covarmat_out_file.open( covarmat_out_filename );
    }
    std::ofstream fullest_out_file;
    if ( write_full_estimate )
    {
         fullest_out_file.open( fullest_out_filename );
         write_full_estimate_header( fullest_out_file,
                                     pheno_covar_snp,
                                     isPhenoBinary,
                                     covar_name_vec );
    } //write_full_estimate

    outfile << "SNP\talpha_G\talpha_G_se\talpha_G_pval\t"; 
    if ( ! pheno_covar_snp )
    {
          outfile << "gamma\tgamma_se\tgamma_pval\n";
    }
    else
    {
        outfile << "beta_G\tbeta_G_se\tbeta_G_pval";
        outfile << "\tgamma\tgamma_se\tgamma_pval\n";
    }
    cout << endl;
    cout << "Iterating over SNPs " << endl;
    vector<string> error_snp_names;
    vector<string> error_snp_errors;
    for ( size_t snp_iter = 0; snp_iter < NUM_SNPS; snp_iter++ )
    {
        if ( (snp_iter % out_freq) == 0 )
        {
            cout << "snp_iter == " << snp_iter << endl;
        }
        const VectorXd snp_vxd = snp_vxd_vec[ snp_iter ];
        MatrixXd Z_mat; 
        make_mxd_with_constant_and_snp_from_covar_mxd( covar_mxd, snp_vxd, Z_mat );

        if ( pheno_covar_snp )
        {
            X_mat = Z_mat;
        }
        double estimate_gamma;
        double estimate_sigma_sqr;
        double estimate_tau_sqr;
        VectorXd estimate_beta;
        VectorXd estimate_ksi;
        int perform_EM_rv;
        VectorXd se_vec;
        MatrixXd covariance_mat;
        VectorXd full_estimate_vec;
        
        if ( false == isPhenoBinary )
        {
           perform_EM_rv = lower_detect_limit::perform_EM( 
                         estimate_gamma, estimate_sigma_sqr,
                         estimate_tau_sqr,
                         estimate_beta, estimate_ksi,
                         Y_pheno_vxd,
                         X_mat,
                         Z_mat,
                         biomarker_measurements,
                         se_vec,
                         max_iter,
                         converged_ratio,
                         covariance_mat,
                         full_estimate_vec
                         );
        }
        else
        {
           perform_EM_rv = 
                 ldl_logistic::perform_EM(
                         estimate_gamma,
                         estimate_tau_sqr,
                         estimate_beta, estimate_ksi,
                         Y_pheno_vxd,
                         X_mat,
                         Z_mat,
                         biomarker_measurements,
                         se_vec,
                         max_iter,
                         converged_ratio,
                         covariance_mat,
                         full_estimate_vec
                         );
        }
        if ( perform_EM_rv != NO_ERROR )
        {
            string my_error_str;
            int_to_error_string( perform_EM_rv, my_error_str );
            cout << "error : For " << snp_name_vector[snp_iter] << 
                     " perform_EM() method ";
            cout << "returned " << my_error_str << endl;
            cout << "Skipping this SNP in output files."<< std::endl;
            error_snp_names.push_back( snp_name_vector[snp_iter] ); 
            error_snp_errors.push_back( my_error_str );  
            continue; //go to next SNP
        }

        if (( estimate_ksi.rows()  < 2 ) ||
            ( estimate_beta.rows() < 2 ))
        {
            string my_error_str;
            int_to_error_string( DIM_ERROR, my_error_str );
            error_snp_names.push_back( snp_name_vector[snp_iter] );
            error_snp_errors.push_back( my_error_str );
            cout << "Error: Skipping this SNP in output files." << endl;
            if (  estimate_ksi.rows()  < 2  )
            {

               cout << "error: estimate_ksi.rows() == "
                    <<   estimate_ksi.rows()
                    << endl;
            }
            if ( estimate_beta.rows() < 2 )
            {
                cout << "error: estimate_beta.rows() == "
                    <<   estimate_beta.rows()
                    << endl;
            }
            continue;
		} //ksi or beta estimate don't both have at least 2 elements
		
        const size_t ksi_second_component_index = 
                estimate_beta.rows() + 3;
		const size_t beta_second_component_index = 1;
		const size_t gamma_index = estimate_beta.rows();
        if (  ( se_vec.rows() < (ksi_second_component_index + 1 ) ) ||
              ( se_vec.rows() < (beta_second_component_index + 1) ) || 
              ( se_vec.rows() < (gamma_index + 1 ) ) ) 
        {
            string my_error_str;
            int_to_error_string( DIM_ERROR, my_error_str );
            error_snp_names.push_back( snp_name_vector[snp_iter] );
            error_snp_errors.push_back( my_error_str );
            cout << "Error: Skipping this SNP in output files." << endl;
            cout << "error: se_veec.rows() == " << se_vec.rows() << endl;
            cout << "se_vec.rows() should be at least as large as " <<
                    "all the values below: " << endl;
            cout << "(ksi_second_component_index + 1 ) == " <<
                     (ksi_second_component_index + 1 ) << endl;
            cout << "(beta_second_component_index + 1 ) == " << 
                     (beta_second_component_index + 1 ) << endl;
            cout << "(gamma_index + 1 ) == " << 
                     (gamma_index + 1 ) << endl;           

            continue;
		} //se_vec does not have enough elements
		
        if ( write_covariance )
        {
           if ( perform_EM_rv == NO_ERROR )
           {
            covarmat_out_file << snp_name_vector[snp_iter] << endl;
            covarmat_out_file << "beta_constant" << "\t";
            if ( pheno_covar_snp )
            {
                covarmat_out_file << "beta_G" << "\t";
            }
            for ( const auto & covar_name : covar_name_vec )
            {
                covarmat_out_file << "beta_" << covar_name 
                               << "\t";
            }
            covarmat_out_file << "gamma" << "\t";
            if ( false == isPhenoBinary )
            {
                covarmat_out_file << "sigma_sqr" << "\t";
            }
            covarmat_out_file << "alpha_constant" << "\t";
            covarmat_out_file << "alpha_G" << "\t";
            for ( const auto & covar_name : covar_name_vec )
            {
                covarmat_out_file << "alpha_" << covar_name 
                               << "\t";
            }
            covarmat_out_file << "tau_sqr" << endl;
            covarmat_out_file << covariance_mat << endl << endl;
          }
        }
        if ( write_full_estimate )
        {
            fullest_out_file << snp_name_vector[snp_iter] << "\t";
            const size_t NUM_COOR_X = estimate_beta.rows();    
			const size_t NUM_COOR_Z = estimate_ksi.rows();
            
            write_full_estimate_data( fullest_out_file,
                                      full_estimate_vec,
                                      se_vec,           
                                      NUM_COOR_X,          
                                      NUM_COOR_Z,            
                                      estimate_gamma,
                                      estimate_sigma_sqr,
                                      estimate_tau_sqr,
                                      isPhenoBinary );

        } //write_full_estimate

        VectorXd estimates_of_interest = VectorXd::Zero(3);
        estimates_of_interest( 0 ) = estimate_ksi( 1 );
        estimates_of_interest( 1 ) = estimate_gamma;
        estimates_of_interest( 2 ) = estimate_beta( 1 );
        VectorXd se_of_interest = VectorXd::Zero( 3 );
        se_of_interest( 0 ) = se_vec( ksi_second_component_index );
        se_of_interest( 1 ) = se_vec( gamma_index );
        se_of_interest( 2 ) = se_vec( beta_second_component_index );
        VectorXd z_stat_vxd;
        calc_z_stat_vec( estimates_of_interest, se_of_interest, z_stat_vxd );
        VectorXd p_val_vxd;
        calc_p_val_vec( z_stat_vxd, p_val_vxd );
        const double ksi_p_val = p_val_vxd( 0 );
        const double gamma_p_val = p_val_vxd( 1 );
        const double beta_p_val = p_val_vxd( 2 );
        outfile << snp_name_vector[ snp_iter ];
        outfile << "\t";
        outfile << estimates_of_interest( 0 );
        outfile << "\t";
        outfile << se_of_interest( 0 );
        outfile << "\t";
        outfile << ksi_p_val;
        outfile << "\t";
        if ( ! pheno_covar_snp )
        {
            outfile << estimates_of_interest( 1 );
            outfile << "\t";
            outfile << se_of_interest( 1 );
            outfile << "\t";
            outfile << gamma_p_val;
            outfile << "\n";
        }
        else
        {
            outfile << estimates_of_interest( 2 );
            outfile << "\t";
            outfile << se_of_interest( 2 );
            outfile << "\t";
            outfile << beta_p_val;
            outfile << "\t";
            outfile << estimates_of_interest( 1 );
            outfile << "\t";
            outfile << se_of_interest( 1 );
            outfile << "\t";                     
            outfile << gamma_p_val; 
            outfile << "\n";
        }
    } //for 
    outfile.close();
    std::cout << "Finished with all SNPs. ";
    if ( error_snp_names.size() != 0 )
    {
        cout << "The following SNPs produced errors "
             << "and were skipped in output files." << endl;;
        
        for ( size_t index =0 ; index < error_snp_names.size(); ++index )
        {
             cout << "SNP: " << error_snp_names[index] << "\t";
             cout << error_snp_errors[index] << endl;
        }
    }
    cout << "Results written to " << out_filename << std::endl;
    if ( write_covariance )
    {
        covarmat_out_file.close();
        cout << "Covariance matrix for each SNP written to " << covarmat_out_filename 
             << endl;
    }
    if ( write_full_estimate )
    {
        fullest_out_file.close();
        cout << "Full estimate for each SNP written to "
             << fullest_out_filename << endl;
    }

    cout << "Done." << endl;
    return EXIT_SUCCESS;
}
